@@ -1,11 +1,14 @@
 import csv
 import json
 import pendulum
+import tempfile
+
 import singer
 from singer import metadata
-from tap_marketo.client import ExportFailed
 from singer import bookmarks
 
+from tap_marketo.client import ExportFailed
+
 # We can request up to 30 days worth of activities per export.
 MAX_EXPORT_DAYS = 30
 
@@ -32,6 +35,7 @@ ITER_CHUNK_SIZE = 512
 
 ATTRIBUTION_WINDOW_DAYS = 1
 
+
 def format_value(value, schema):
     if not isinstance(schema["type"], list):
         field_type = [schema["type"]]
@@ -64,6 +68,113 @@ def format_values(stream, row):
         rtn[field] = format_value(row.get(field), schema)
     return rtn
 
+
+def update_state_with_export_info(state, stream, bookmark=None, export_id=None, export_end=None):
+    state = bookmarks.write_bookmark(state, stream["tap_stream_id"], "export_id", export_id)
+    state = bookmarks.write_bookmark(state, stream["tap_stream_id"], "export_end", export_end)
+    if bookmark:
+        state = bookmarks.write_bookmark(state, stream["tap_stream_id"], stream["replication_key"], bookmark)
+
+    singer.write_state(state)
+    return state
+
+
+def get_export_end(export_start):
+    export_end = export_start.add(days=MAX_EXPORT_DAYS)
+    if export_end >= pendulum.utcnow():
+        export_end = pendulum.utcnow()
+
+    return export_end.replace(microsecond=0)
+
+
+def wait_for_export(client, state, stream, export_id):
+    stream_type = "activities" if stream["tap_stream_id"] != "leads" else "leads"
+    try:
+        client.wait_for_export(stream_type, export_id)
+    except ExportFailed:
+        state = update_state_with_export_info(state, stream)
+        raise
+
+    return state
+
+
+def stream_rows(client, stream_type, export_id):
+    with tempfile.NamedTemporaryFile(mode="w+", encoding="utf8") as csv_file:
+        singer.log_info("Download starting.")
+        resp = client.stream_export(stream_type, export_id)
+        for chunk in resp.iter_content(chunk_size=1024, decode_unicode=True):
+            if chunk:
+                csv_file.write(chunk)
+
+        singer.log_info("Download completed. Begin streaming rows.")
+        csv_file.seek(0)
+        reader = csv.reader(csv_file, delimiter=',', quotechar='"')
+        headers = next(reader)
+        for line in reader:
+            yield dict(zip(headers, line))
+
+
+def get_or_create_export_for_leads(client, state, stream, export_start):
+    export_id = bookmarks.get_bookmark(state, "leads", "export_id")
+    # check if export is still valid
+    if not client.export_available("leads", export_id):
+        singer.log_info("Export %s no longer available.", export_id)
+        export_id = None
+
+    if export_id is None:
+        # Corona mode is required to query by "updatedAt", otherwise a full
+        # sync is required using "createdAt".
+        query_field = "updatedAt" if client.use_corona else "createdAt"
+        export_end = get_export_end(export_start)
+        query = {query_field: {"startAt": export_start.isoformat(),
+                               "endAt": export_end.isoformat()}}
+
+        # Create the new export and store the id and end date in state.
+        # Does not start the export (must POST to the "enqueue" endpoint).
+        fields = [f for f, s in stream["schema"]["properties"].items()
+                  if s.get("selected") or (s.get("inclusion") == "automatic")]
+        export_id = client.create_export("leads", fields, query)
+        state = update_state_with_export_info(
+            state, stream, export_id=export_id, export_end=export_end.isoformat())
+    else:
+        export_end = pendulum.parse(bookmarks.get_bookmark(state, "leads", "export_end"))
+
+    return export_id, export_end
+
+
+def get_or_create_export_for_activities(client, state, stream, export_start):
+    export_id = bookmarks.get_bookmark(state, stream["tap_stream_id"], "export_id")
+    if not client.export_available(stream["tap_stream_id"], export_id):
+        singer.log_info("Export %s no longer available.", export_id)
+        export_id = None
+
+    if export_id is None:
+        # The activity id is in the top-most breadcrumb of the metatdata
+        # Activity ids correspond to activity type id in Marketo.
+        # We need the activity type id to build the query.
+        activity_metadata = metadata.to_map(stream["metadata"])
+        activity_type_id = metadata.get(activity_metadata, (), 'marketo.activity-id')
+
+        # Activities must be queried by `createdAt` even though
+        # that is not a real field. `createdAt` proxies `activityDate`.
+        # The activity type id must also be included in the query. The
+        # largest date range that can be used for activities is 30 days.
+        export_end = get_export_end(export_start)
+        query = {"createdAt": {"startAt": export_start.isoformat(),
+                               "endAt": export_end.isoformat()},
+                 "activityTypeIds": [activity_type_id]}
+
+        # Create the new export and store the id and end date in state.
+        # Does not start the export (must POST to the "enqueue" endpoint).
+        export_id = client.create_export("activities", ACTIVITY_FIELDS, query)
+        state = update_state_with_export_info(
+            state, stream, export_id=export_id, export_end=export_end.isoformat())
+    else:
+        export_end = pendulum.parse(bookmarks.get_bookmark(state, stream["tap_stream_id"], "export_end"))
+
+    return export_id, export_end
+
+
 def flatten_activity(row, stream):
     # Start with the base fields
     rtn = {field: row[field] for field in BASE_ACTIVITY_FIELDS}
@@ -87,235 +198,58 @@ def flatten_activity(row, stream):
 
     return rtn
 
-def get_or_create_export_for_leads(client, state, stream, fields, start_date):
-    export_id = bookmarks.get_bookmark(state, stream["tap_stream_id"], "export_id")
-    export_end = bookmarks.get_bookmark(state, stream["tap_stream_id"], "export_end")
-
-    if export_id is None:
-        if client.use_corona:
-            query_field = "updatedAt"
-        else:
-            query_field = "createdAt"
-        
-        export_end = start_date.add(days=MAX_EXPORT_DAYS)
-        if export_end >= pendulum.utcnow():
-            export_end = pendulum.utcnow()
-
-        export_end = export_end.replace(microsecond=0)
-            
-        query = {query_field: {"startAt": start_date.isoformat(), "endAt": export_end.isoformat()}}
-        # Create the new export and store the id and end date in state.
-        # Does not start the export (must POST to the "enqueue" endpoint).
-        export_id = client.create_export("leads", fields, query)
-        update_state_with_export_info(state, stream, export_id=export_id, export_end=export_end.isoformat())
-
-    else:
-        export_end = pendulum.parse(export_end)    
-
-    return export_id, export_end
-
-def _iter_lines(response_lines):
-    """Clone of the iter_lines function from the requests library with the change
-        to pass keepends=True in order to ensure that we do not strip the line breaks
-        from within a quoted value from the CSV stream."""
-    pending = None
-
-    for chunk in response_lines.iter_content(decode_unicode=True, chunk_size=ITER_CHUNK_SIZE):
-        if pending is not None:
-            chunk = pending + chunk
-
-        lines = chunk.splitlines(keepends=True)
-        
-        if lines and lines[-1] and chunk and lines[-1][-1] == chunk[-1]:
-            pending = lines.pop()
-        else:
-            pending = None
-
-        for line in lines:
-            yield line
-
-    if pending is not None:
-        yield pending
-
-def write_leads_records(client, stream, lines, og_bookmark_value, record_count):
-    null_updatedAt_count = 0
-
-    csv_stream = csv.reader(_iter_lines(lines),
-                            delimiter=',',
-                            quotechar='"')
-
-    headers = next(csv_stream)
-
-    for line in csv_stream:        
-        line = dict(zip(headers, line))
-
-        #deal with updatedAt potentially being null
-        line_updated_at = line.get('updatedAt')
-        if line_updated_at == 'null' or line_updated_at is None:
-            null_updatedAt_count += 1
-            if null_updatedAt_count <= 10:
-                singer.log_info("Found record with null updatedAt value.  The id value for the record is %s", line.get('id'))    
-        else:
-            line_updated_at = pendulum.parse(line["updatedAt"])
-
-        #accounts without corona need to have a manual filter on records
-        if client.use_corona or line_updated_at == 'null' or (line_updated_at >= og_bookmark_value):
-            record = format_values(stream, line)
-            singer.write_record(stream["tap_stream_id"], record)
-            record_count += 1
-
-    if null_updatedAt_count > 0:
-        singer.log_info("For this export: Count of null updatedAt fields: %d", null_updatedAt_count)
-    return record_count
 
 def sync_leads(client, state, stream):
     # http://developers.marketo.com/rest-api/bulk-extract/bulk-lead-extract/
-    singer.write_schema(stream["tap_stream_id"], stream["schema"], stream["key_properties"])
+    singer.write_schema("leads", stream["schema"], stream["key_properties"])
+    initial_bookmark = pendulum.parse(bookmarks.get_bookmark(state, "leads", stream["replication_key"]))
+    export_start = pendulum.parse(bookmarks.get_bookmark(state, "leads", stream["replication_key"]))
+    if client.use_corona:
+        export_start = export_start.subtract(days=ATTRIBUTION_WINDOW_DAYS)
+
+    job_started = pendulum.utcnow()
     record_count = 0
-    replication_key = stream.get("replication_key")
-    tap_stream_id = stream.get("tap_stream_id")
-    tap_job_start_time = pendulum.utcnow()
-    
-    fields = [f for f, s in stream["schema"]["properties"].items() if s.get("selected") or (s.get("inclusion") == "automatic")]
+    max_bookmark = initial_bookmark
+    while export_start < job_started:
+        export_id, export_end = get_or_create_export_for_leads(client, state, stream, export_start)
+        state = wait_for_export(client, state, stream, export_id)
+        for row in stream_rows(client, "leads", export_id):
+            record = format_values(stream, row)
+            record_bookmark = pendulum.parse(record["updatedAt"])
 
-    og_bookmark_value = pendulum.parse(bookmarks.get_bookmark(state, tap_stream_id, replication_key))
-    bookmark_date = og_bookmark_value.subtract(days=ATTRIBUTION_WINDOW_DAYS)
-
-    while bookmark_date < tap_job_start_time:
-        export_id, export_end = get_or_create_export_for_leads(client, state, stream, fields, bookmark_date)
-
-        try:
-            client.wait_for_export("leads", export_id)
-        except ExportFailed as ex:
-            update_state_with_export_info(state, stream)
-            singer.log_critical("Export job failure.  Status was" + ex)
-
-        lines = client.stream_export("leads", export_id)
-
-        try:
-            record_count = write_leads_records(client, stream, lines, og_bookmark_value, record_count)
-
-        except Exception as e:
-            singer.log_info("Exception while writing leads record, removing export information from state")
-            update_state_with_export_info(state, stream)
-            raise e
-
-        if client.use_corona:
-            state = update_state_with_export_info(state, stream, bookmark=export_end.isoformat(), \
-                                          export_id=None, export_end=None)
-        else:
-            state = update_state_with_export_info(state, stream, \
-                                          export_id=None, export_end=None)
-
-        bookmark_date = export_end
-
-    state = update_state_with_export_info(state, stream, bookmark=bookmark_date.isoformat(), \
-                                          export_id=None, export_end=None)
+            if client.use_corona:
+                max_bookmark = export_end
+                singer.write_record("leads", record)
+                record_count += 1
+            elif record_bookmark >= initial_bookmark:
+                max_bookmark = max(max_bookmark, record_bookmark)
+                singer.write_record("leads", record)
+                record_count += 1
 
+        # Now that one of the exports is finished, update the bookmark
+        state = update_state_with_export_info(state, stream, bookmark=max_bookmark.isoformat())
+        export_start = export_end
 
     return state, record_count
 
-def get_or_create_export_for_activities(client, state, stream):
-    export_id = bookmarks.get_bookmark(state, stream["tap_stream_id"], "export_id")
-
-    if not export_id:
-        # The activity id is in the top-most breadcrumb of the metatdata
-        # Activity ids correspond to activity type id in Marketo.
-        # We need the activity type id to build the query.
-        activity_metadata = metadata.to_map(stream["metadata"])
-        activity_type_id = metadata.get(activity_metadata, (), 'marketo.activity-id')
-        singer.log_info("activity id for stream %s is %d", stream["tap_stream_id"], activity_type_id)
-        
-        # Activities must be queried by `createdAt` even though
-        # that is not a real field. `createdAt` proxies `activityDate`.
-        # The activity type id must also be included in the query. The
-        # largest date range that can be used for activities is 30 days.
-        start_date = bookmarks.get_bookmark(state, stream["tap_stream_id"], stream["replication_key"])
-        start_pen = pendulum.parse(start_date)
-        end_pen = start_pen.add(days=MAX_EXPORT_DAYS)
-        if end_pen >= pendulum.utcnow():
-            end_pen = pendulum.utcnow()
-        end_pen = end_pen.replace(microsecond=0)
-        end_date = end_pen.isoformat()
-
-        query = {"createdAt": {"startAt": start_pen.isoformat(), "endAt": end_date},
-                 "activityTypeIds": [activity_type_id]}
-        singer.log_info("scheduling export for stream \"%s\" with query: %s", stream["tap_stream_id"], query)
-        # Create the new export and store the id and end date in state.
-        # Does not start the export (must POST to the "enqueue" endpoint).
-        export_id = client.create_export("activities", ACTIVITY_FIELDS, query)
-        update_state_with_export_info(state, stream, export_id=export_id, export_end=end_date)
-
-    return export_id
-
-
-def update_state_with_export_info(state, stream, bookmark=None, export_id=None, export_end=None):
-    state = bookmarks.write_bookmark(state, stream["tap_stream_id"], "export_id", export_id)
-    state = bookmarks.write_bookmark(state, stream["tap_stream_id"], "export_end", export_end)
-    if bookmark:
-        state = bookmarks.write_bookmark(state, stream["tap_stream_id"], stream["replication_key"], bookmark)
-
-    singer.write_state(state)
-    return state
-
-def convert_line(stream, headers, line):
-    row = dict(zip(headers, line))
-    row = flatten_activity(row, stream)
-    return format_values(stream, row)
-
-
-def handle_record(state, stream, record):
-    start_date = bookmarks.get_bookmark(state, stream["tap_stream_id"], stream["replication_key"])
-    if record[stream["replication_key"]] < start_date:
-        return 0
-
-    singer.write_record(stream["tap_stream_id"], record)
-    return 1
-
-
-def wait_for_activity_export(client, state, stream, export_id):
-    try:
-        client.wait_for_export("activities", export_id)
-    except ExportFailed:
-        update_state_with_export_info(state, stream)
-        raise
-
 
 def sync_activities(client, state, stream):
     # http://developers.marketo.com/rest-api/bulk-extract/bulk-activity-extract/
     singer.write_schema(stream["tap_stream_id"], stream["schema"], stream["key_properties"])
-    start_date = bookmarks.get_bookmark(state, stream["tap_stream_id"], stream["replication_key"])
-    start_pen = pendulum.parse(start_date)
+    export_start = pendulum.parse(bookmarks.get_bookmark(state, stream["tap_stream_id"], stream["replication_key"]))
     job_started = pendulum.utcnow()
     record_count = 0
+    while export_start < job_started:
+        export_id, export_end = get_or_create_export_for_activities(client, state, stream, export_start)
+        state = wait_for_export(client, state, stream, export_id)
+        for row in stream_rows(client, "activites", export_id):
+            row = flatten_activity(row, stream)
+            record = format_value(stream, row)
+            singer.write_record(stream["tap_stream_id"], record)
+            record_count += 1
 
-    while start_pen < job_started:
-        export_id = get_or_create_export_for_activities(client, state, stream)
-
-        # If the export fails while running, clear the export information
-        # from state so a new export can be run next sync.
-        wait_for_activity_export(client, state, stream, export_id)
-
-        try:
-            resp = client.stream_export("activities", export_id)
-            csv_stream = csv.reader(_iter_lines(resp),
-                                    delimiter=',',
-                                    quotechar='"')
-            headers = next(csv_stream)
-            for line in csv_stream:        
-                record = convert_line(stream, headers, line)
-                record_count += handle_record(state, stream, record)
-        except Exception as e:
-            singer.log_info("Exception while writing activity \"%s\" record, removing export information from state", stream["tap_stream_id"])
-            update_state_with_export_info(state, stream)
-            raise e            
-                
-        # The new start date is the end of the previous export. Update
-        # the bookmark to the end date and continue with the next export.
-
-        start_date = bookmarks.get_bookmark(state, stream["tap_stream_id"], "export_end")
-        update_state_with_export_info(state, stream, bookmark=start_date)
-        start_pen = pendulum.parse(start_date)
+        state = update_state_with_export_info(state, stream, bookmark=export_start.isoformat())
+        export_start = export_end
 
     return state, record_count
 
@@ -488,10 +422,6 @@ def sync(client, catalog, state):
 
     # If Corona is not supported, log a warning near the end of the tap
     # log with instructions on how to get Corona supported.
-    singer.log_info("Performing final Corona check")
+    singer.log_info("Finished sync.")
     if not client.use_corona:
-        singer.log_info("Finished sync")        
         singer.log_warning(NO_CORONA_WARNING)
-    else:
-        singer.log_info("Finished sync")        
-        
